"""
A built in entity model instance for adding a device to be considered as part of the IOC startup process.
This allows the IOC startup process to pause until communication with the specified devices is established,
which can help prevent issues with devices not being ready when the IOC starts.
(e.g used to detect motion controller on the network or mounted usb devices)
"""

from pathlib import Path
from typing import Literal

from pydantic import Field
from ruamel.yaml import YAML

from ibek.globals import GLOBALS
from ibek.ioc import BuiltInEntity

WAIT4IP_TYPE = "ibek.wait_ip"
WAIT4USB_TYPE = "ibek.wait_usb"


class DoWaitEntity(BuiltInEntity):
    """
    A definition of DoWaitEntity for the type checker.

    This is not really used - instead the dynamic class created
    by the make_entity_models function is used.
    """

    type: Literal["ibek.do_wait"] = "ibek.do_wait"
    device: str = Field(
        description="The device name to use in the database record.",
        default="DEVICE",
    )
    timeout: int = Field(
        description="The number of seconds to wait for a response before considering the communication attempt a failure "
        "and exiting the IOC startup process; this will trigger a restart of the IOC pod by Kubernetes.\n"
        "A value of 0 means to wait indefinitely until communication is established.",
        default=10000,
    )

    def _process_entity(self, output: Path = GLOBALS.RUNTIME_OUTPUT):
        pass


class Wait4USBEntity(DoWaitEntity):
    """
    A definition of Wait4USBEntity for the type checker.

    This is not really used - instead the dynamic class created
    by the make_entity_models function is used.
    """

    type: Literal["ibek.wait_usb"] = "ibek.wait_usb"  # type: ignore[assignment]
    id: str = Field(
        description="The ID of the USB device to wait for."
        "This should be in the format 'vendor_id:product_id', where vendor_id and product_id are the hexadecimal IDs of the USB device."
        "For example, '1234:5678'.",
    )

    def _process_entity(self, output: Path = GLOBALS.RUNTIME_OUTPUT):
        """
        Not implemented yet - this will likely involve writing the USB device ID to a file in the runtime directory,
        for external tooling to parse and use to detect when the device is present.
        """
        return super()._process_entity(output)


class Wait4IPEntity(DoWaitEntity):
    """
    A definition of Wait4IPEntity for the type checker.

    This is not really used - instead the dynamic class is created
    by the make_entity_models function is used.
    """

    type: Literal["ibek.wait_ip"] = "ibek.wait_ip"  # type: ignore[assignment]
    address: str = Field(
        description="The IP address and port to check communication with.",
    )

    def _process_entity(self, output: Path = GLOBALS.RUNTIME_OUTPUT):
        """
        Write the entity parameters into a YAML list file under the runtime directory.
        The file is created if it doesn't already exist and any previous entries are preserved.
        This makes it easy for external tooling to parse the list of addresses that need waiting for.
        """
        # Make sure the runtime directory exists
        output.mkdir(parents=True, exist_ok=True)

        # Create the YAML file path and initialize it with a header if it doesn't exist
        yaml_path = output / "wait_list.yaml"
        yaml_path.parent.mkdir(parents=True, exist_ok=True)

        yaml = YAML()
        header_comment = (
            "#######################################################################################\n"
            "# List of hardware to wait for communication with before proceeding with the IOC start.\n"
            "# This file is generated by Wait4IPEntity._process_entity_no_file.\n"
            "#######################################################################################\n\n"
        )

        if yaml_path.exists():
            header_needed = False
            # Load existing data from the YAML file, or initialize an empty list if the file is empty
            data = yaml.load(yaml_path) or []
        else:
            header_needed = True
            data = []

        # Define the entry for this entity and append it to the data list
        entry = {
            "type": self.type,
            "device": self.device,
            "address": self.address,
            "timeout": self.timeout,
        }
        data.append(entry)

        # Write the updated data back to the YAML file, preserving any existing entries and adding a header if the file was newly created
        with yaml_path.open("w") as stream:
            if header_needed:
                stream.write(header_comment)
            yaml.dump(data, stream)
            stream.write("\n")  # newline at eof for better readability



import os
import re
import subprocess
from enum import StrEnum
from pathlib import Path
from typing import Annotated

import typer

from ibek.globals import (
    GLOBALS,
    NaturalOrderGroup,
)
from ibek.support import Support


class AptWhen(StrEnum):
    dev = "dev"
    run = "run"
    both = "both"


# find macro name and macro value in a RELEASE file
# only include values with at least one / (attempt to match filepaths only)
PARSE_MACROS = re.compile(r"^([A-Z_a-z0-9]*)\s*=\s*(.*/.*)$", flags=re.M)


support_cli = typer.Typer(cls=NaturalOrderGroup)


def _install_debs(debs: list[str]) -> None:
    """
    Install a list of debian packages.

    If they have an http:// or https://
    prefix then they will be downloaded and installed from file.

    args: debs: List[str] - list of debian packages to install - can also include
                            any additional 'apt-get install' options required
    """
    temp = Path("/tmp")
    for i, pkg in enumerate(debs):
        if pkg.startswith("http://") or pkg.startswith("https://"):
            pkg_file = temp / pkg.split("/")[-1]
            subprocess.call(["busybox", "wget", pkg, "-O", str(pkg_file)])
            debs[i] = str(pkg_file)

    if len(debs) == 0:
        print("no packages to install")
        return

    print("installing packages: ", debs)

    sudo = "sudo" if os.geteuid() != 0 else ""
    command = (
        f"{sudo} apt-get update && {sudo} apt-get upgrade -y && "
        f"{sudo} apt-get install -y --no-install-recommends "
        + " ".join(debs)
        + f" && {sudo} rm -rf /var/lib/apt/lists/*"
    )
    exit(subprocess.call(["bash", "-c", command]))


# TODO this should be deprecated in favor of an additional Ansible role
@support_cli.command()
def apt_install_runtime_packages():
    """
    Install packages from the list collected by calls to add_runtime_packages
    """
    if GLOBALS.RUNTIME_DEBS.exists():
        debs = GLOBALS.RUNTIME_DEBS.read_text().split()
        _install_debs(debs)


@support_cli.command()
def generate_schema(
    output: Annotated[
        Path | None,
        typer.Option(
            help="The filename to write the schema to",
            autocompletion=lambda: [],  # Forces path autocompletion
        ),
    ] = None,
):
    """Produce JSON global schema for all <support_module>.ibek.support.yaml files"""
    if output is None:
        typer.echo(Support.get_schema())
    else:
        output.write_text(Support.get_schema() + "\n")



"""
System tests for the ibek builtin commands
"""

from pathlib import Path

from tests.test_cli import generic_generate

# Note that the tests will fail unless the './generate_samples.sh' script has been run first
# to generate the expected test output in the './samples/outputs/wait' directory.


def test_wait_example(tmp_epics_root: Path, samples):
    generic_generate(
        tmp_epics_root,
        samples,
        "wait",
        ["epics", "asyn", "motorSim"],
    )



"""
functions for making Entity models
"""

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Annotated, Union

from pydantic import Field
from ruamel.yaml.main import YAML

from ibek.parameters import EnumParam
from ibek.utils import UTILS

from .entity_model import EntityModel
from .ioc import IOC


class IocFactory:
    """
    A Class for creating Generic IOC classes and instances
    """

    def deserialize_ioc(
        self, ioc_instance_yaml: Path, entity_models: list[EntityModel]
    ) -> IOC:
        """
        Takes an ioc instance entities file, list of generic ioc yaml files.

        Returns a model of the resulting ioc instance
        """

        # Create a Class representing this Generic IOC
        ioc_model = self.make_ioc_model(entity_models)

        # extract the ioc instance yaml into a dict
        ioc_instance_dict = YAML(typ="safe").load(ioc_instance_yaml)

        if ioc_instance_dict is None or "ioc_name" not in ioc_instance_dict:
            raise RuntimeError(
                f"Failed to load a valid ioc config from {ioc_instance_yaml}"
            )

        # extract the ioc name into UTILS for use in jinja renders
        name = UTILS.render({}, ioc_instance_dict["ioc_name"], "str")
        UTILS.set_ioc_name(name)
        ioc_instance_dict["ioc_name"] = name

        # Create an IOC instance from the instance dict and the model
        ioc_instance = ioc_model(**ioc_instance_dict)
        self.fixup_enums(ioc_instance)

        return ioc_instance

    def fixup_enums(self, ioc_instance: IOC):
        """
        Fixup the enums in the IOC instance, so that they are the value of
        their original Enum, rather than the key.

        *.ibek.support.yaml files may specify Enum parameters like this:
        ```
            stop:
              type: enum
              description: |-
              Stop Bits
              values:
                one: 1
                two: 2
              default: "1"
        ```
        This means that ioc.yaml would specify stop bits using the strings
        "one" and "two", but when the startup script or DB renders the value
        it will use the Enum value, which is 1 and 2 respectively.

        To do this requires the use of "use_enum_values=True" in the Pydantic
        model config so that only a string need be specified instead of a
        serialized Enum.

        To make this work we swap the value and key in all enums before creating
        the model (see enum_swapped in EntityFactory._make_entity_type). Pydantic
        will then serialize/deserialize using the value of the swapped Enum
        i.e. "one" or "two".

        After deserializing the IOC instance this function swaps back to the
        original value so that it will be used in rendering of parameters. Enums
        can also be specified with no value, in which case the key is used
        in ioc.yaml and in rendering.
        """
        for entity in ioc_instance.entities:
            for field_name, field_value in entity.model_dump().items():
                # RepeatEntities are dumb dictionaries with no _model, skip them
                if hasattr(entity, "_model"):
                    # check to see if field is an Enum value in the EntityModel
                    parameter = entity._model.parameters.get(field_name)
                    if isinstance(parameter, EnumParam):
                        # if so, set the value to the Enum value, instead of its key
                        enum_value = parameter.values.get(field_value)
                        if enum_value is not None:
                            setattr(entity, field_name, enum_value)

    def make_ioc_model(self, entity_models: list[EntityModel]) -> type[IOC]:
        """
        Create an IOC derived model, by setting its entities attribute to
        be of type 'list of Entity derived classes'.

        Also instantiate any SubEntities at this time
        """

        entity_union = Union[tuple(entity_models)]  # type: ignore
        discriminated = Annotated[entity_union, Field(discriminator="type")]  # type: ignore

        class NewIOC(IOC):
            entities: Sequence[discriminated] = Field(  # type: ignore
                description="List of entities this IOC instantiates"
            )

        return NewIOC



"""
System tests that run the CLI commands and compare the output to expected
results.
"""

import json
import subprocess
import sys
from pathlib import Path

import pytest
from typer.testing import CliRunner

import ibek.utils as utils
from ibek import __version__
from ibek.runtime_cmds.commands import do_generate
from tests.conftest import run_cli

runner = CliRunner()


def test_cli_version():
    cmd = [sys.executable, "-m", "ibek", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_ibek_schema(tmp_path: Path, samples: Path):
    """generate the global ibek schema"""
    schema_path = tmp_path / "schema.json"
    run_cli("support", "generate-schema", "--output", schema_path)
    expected = json.loads(
        (samples / "schemas" / "ibek.support.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_single_schema(tmp_path: Path, samples: Path):
    """generate schema from a support module definition yaml"""

    schema_path = tmp_path / "single.ibek.support.schema.json"
    yaml_path = samples / "support" / "motorSim.ibek.support.yaml"
    run_cli(
        "ioc", "generate-schema", "--no-ibek-defs", yaml_path, "--output", schema_path
    )

    expected = json.loads(
        (samples / "schemas" / "single.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_motor_sim_schema(tmp_path: Path, samples: Path):
    """generate schema for a container with two support modules"""

    schema_combined = tmp_path / "motorSim.ibek.ioc.schema.json"
    yaml_path1 = samples / "support" / "asyn.ibek.support.yaml"
    yaml_path2 = samples / "support" / "motorSim.ibek.support.yaml"
    run_cli(
        "ioc",
        "generate-schema",
        "--no-ibek-defs",
        yaml_path1,
        yaml_path2,
        "--output",
        schema_combined,
    )

    expected = json.loads(
        (samples / "schemas" / "motorSim.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_combined).read_text())
    assert expected == actual


def test_build_runtime_motorSim(tmp_epics_root: Path, samples: Path):
    """
    build an ioc runtime script from an IOC instance entity file
    and multiple support module definition files

    Also verifies database subst file generation for multiple
    entity instantiations.
    """

    generic_generate(
        tmp_epics_root,
        samples,
        "motorSim",
        ["motorSim", "asyn"],
    )


def test_build_utils_features(tmp_epics_root: Path, samples: Path):
    """
    build an ioc runtime script to verify utils features
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "utils",
        ["utils"],
    )


def test_ipac(tmp_epics_root: Path, samples: Path):
    """
    Tests that an id argument can include another argument in its default value
    """

    # reset the InterruptVector counter to its initial state (if already used)
    if "InterruptVector" in utils.UTILS.variables:
        utils.UTILS.variables["InterruptVector"] = 191

    generic_generate(
        tmp_epics_root,
        samples,
        "ipac-test",
        ["ipac", "epics"],
    )


def test_gauges(tmp_epics_root: Path, samples: Path):
    """
    Tests that an id argument can include another argument in its default value
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "gauges",
        ["asyn", "gauges"],
    )


def test_quadem(tmp_epics_root: Path, samples: Path):
    """
    Tests the use of CollectionDefinitions in an IOC instance
    this example uses the tetramm beam position monitor module
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "quadem",
        ["ADCore", "quadem"],
    )


def generic_generate(
    epics_root: Path, samples: Path, ioc_yaml_name: str, support_names: list[str]
):
    ioc_yaml = samples / "iocs" / f"{ioc_yaml_name}.ibek.ioc.yaml"
    support_yamls = [
        samples / "support" / f"{name}.ibek.support.yaml" for name in support_names
    ]
    expected_outputs = samples / "outputs" / ioc_yaml_name

    do_generate([ioc_yaml], support_yamls, epics_root / "runtime", pvi=True)

    outputs = list(expected_outputs.glob("*"))
    assert len(outputs) > 0, "No expected output files found"

    for output in outputs:
        actual = epics_root / "runtime" / output.name
        if not actual.exists():
            actual = epics_root / "opi" / output.name
        assert actual.exists(), f"Missing output file {actual}"
        assert output.read_text().strip() == actual.read_text().strip()


def test_andreas_motors(tmp_epics_root: Path, samples: Path):
    """
    Motor and axis example
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "technosoft",
        ["asyn", "technosoft"],
    )


def test_list(tmp_epics_root: Path, samples: Path):
    """
    Motor and axis example
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "listarg",
        ["listarg"],
    )


def test_fast_vacuum(tmp_epics_root: Path, samples: Path):
    """
    Cut down copy of dlsPLC containing fast vacuum master and channel
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "fastVacuum",
        ["fastVacuum"],
    )


def test_dls_plc(tmp_epics_root: Path, samples: Path):
    """
    Cut down copy of dlsPLC containing fast vacuum master and channel
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "dlsPLC",
        ["dlsPLC", "asyn"],
    )


def test_dls8515(tmp_epics_root: Path, samples: Path):
    """
    Cut down copy of dlsPLC containing fast vacuum master and channel
    """
    with pytest.raises(ValueError) as ctx:
        generic_generate(
            tmp_epics_root,
            samples,
            "DLS8515",
            ["DLS8515", "ipac", "epics"],
        )

    assert "`ipac.Hy8002`.interrupt_vector" in str(ctx.value)



"""
Tests to verify that the error handling works as expected.
"""

from pathlib import Path

import pytest
from ruamel.yaml import YAML

from ibek.ioc_factory import IocFactory
from ibek.runtime_cmds.commands import do_generate
from ibek.support import Support
from ibek.utils import UTILS
from tests.conftest import run_cli


def test_counter_overuse(tmp_epics_root: Path, samples: Path):
    """
    Check that counter limits are enforced
    """
    UTILS.__reset__()

    entity_file = samples / "iocs" / "bad_counter2.ibek.ioc.yaml"
    definition_file1 = samples / "support" / "utils.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        do_generate(
            [entity_file],
            [definition_file1],
            output_folder=tmp_epics_root / "runtime",
            pvi=False,
        )

        run_cli("runtime", "generate", entity_file, definition_file1)

    assert "Counter InterruptVector exceeded maximum value of 194" in str(ctx.value)


def test_bad_ref(samples: Path):
    """
    Check bad object references are caught
    """
    UTILS.__reset__()

    entity_file = samples / "iocs" / "bad_ref.ibek.ioc.yaml"
    definition_file1 = samples / "support" / "asyn.ibek.support.yaml"
    definition_file2 = samples / "support" / "motorSim.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1, definition_file2)
    assert "object controllerOnePort_BAD_REF not found" in str(ctx.value)


def test_bad_db(tmp_epics_root: Path, samples: Path):
    """
    Check bad database args are caught
    """
    UTILS.__reset__()

    entity_file = samples / "iocs" / "bad_db.ibek.ioc.yaml"
    definition_file1 = samples / "support" / "bad_db.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        do_generate(
            [entity_file],
            [definition_file1],
            output_folder=tmp_epics_root / "runtime",
            pvi=False,
        )
    assert "'non-existant' in database template 'test.db' not found" in str(ctx.value)


def test_loading_module_twice(entity_factory, samples: Path):
    """
    Verify we get a sensible error if we try to load a module twice
    without clearing the entity model ids
    """
    definition_file = samples / "support" / "utils.ibek.support.yaml"
    instance_file = samples / "iocs" / "utils.ibek.ioc.yaml"

    support = Support(**YAML(typ="safe").load(definition_file))
    entities1 = entity_factory._make_entity_types(support)
    entities2 = entity_factory._make_entity_types(support)

    ioc_factory = IocFactory()
    generic_ioc1 = ioc_factory.make_ioc_model(entities1)
    generic_ioc2 = ioc_factory.make_ioc_model(entities2)

    instance = YAML(typ="safe").load(instance_file)
    generic_ioc1(**instance)
    with pytest.raises(ValueError) as ctx:
        generic_ioc2(**instance)

    assert "Duplicate id" in str(ctx.value)


def test_defaults(tmp_path: Path, samples: Path):
    """
    Check you cannot redefine a counter with the same name and different params
    """
    entity_file = samples / "iocs" / "bad_default.ibek.ioc.yaml"
    definition_file1 = samples / "support" / "asyn.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1)

    assert "Field required [type=missing" in str(ctx.value)
